# This function is used for all results presented within the main manuscript and supplementary material.

# This function will return parameter estimates and standard deviations of parameter estimates.
# Standard error estimates are outputted for Kauermann and Caroll (2001), 
# Mancl and DeRouen (2001), the average of Kauermann and Carroll (2001) and Mancl and DeRouen (2001),
# Fay and Graubard (2001), Gosho et al. (2014), Morel et al. (2003), and Fan et al. (2013).
# Pan and Wall (2001) df are given for these corresponding SE estimates.
# The function additionally gives Wang and Long (2011) and Westgate and Burchett (2016) df.
# Function inputs are: id - a 1 column matrix identifying what subject number the observation belongs to;
#                      y - a 1 column matrix of outcomes;
#                      x - design matrix not including intercept values;
#                      fam - the family the outcome generates from (gaussian, poisson, binomial)
#                      corstr-the working correlation structure
#                      b - initial parameter estimates, if applicable. For our simulations, we have not specified
#                           initial parameter estimates;
#                      maxiter - the maximum number of iterations allotted to determine convergence
#                      tol - tolerance level for determining convergence
#                      time- a 1 column matrix identifying


GEE=function(id, y, x, fam, corstr, b, maxiter, tol, time)
{
 obs=lapply(split(id,id),"length")
 nobs <- as.numeric(obs)             #vector of cluster sizes for each cluster
 nsub <- length(nobs)                #number of subjects


 max_time = max(time)                #This is the last possible time point




 #Number of summations used for each working correlation structure.  Used when obtaining correlation parameter estimates.

 if (corstr == "exchangeable" )
  {
   N_star = 0
   for(i in 1:nsub)
    N_star = N_star + 0.5*nobs[i]*(nobs[i]-1)  #Used in estimating the correlation
  }

  K_1 = 0 #For AR-1 or MA-1
  Toep_corr=matrix(0,(max_time-1),1) 
  count_un=matrix(0,max_time,max_time) 


  loc1 <- 0
  loc2 <- 0
  for(i in 1:nsub){
   loc1 <- loc2 + 1
   loc2 <- loc1 + nobs[i] - 1
   yi <- as.matrix(y[loc1:loc2, ])
   xi <- x[loc1:loc2, ]
   ni <- nrow(yi)
   ti <- as.matrix(time[loc1:loc2, ])

   if (corstr == "AR-1" | corstr == "MA-1")
   {
    #K_1 = 0
    if(nobs[i]>1)
    {
     for(l in 1:(nobs[i]-1))
      if( (ti[(l+1),1]-ti[l,1]) == 1)
       K_1 = K_1 + 1 
    }
   }


   if (corstr == "Toeplitz") 
   {  
    #Toep_corr=matrix(0,(max_time-1),1) #There are max_time-1 correlation np within the Toeplitz structure. This matrix sums the total number contributed to their estimation.
    if(nobs[i]>1)
    {
     for(k in 1:(max_time-1))
      for(l in 1:(nobs[i]-1))
       for(j in (l+1):nobs[i])
        if( (ti[j,1]-ti[l,1]) == k)
         Toep_corr[k,1] = Toep_corr[k,1] + 1 
    }
   }


   if (corstr == "unstructured") 
   {  
    if(nobs[i]>1)
    {
	for(j in 1:(nobs[i]-1))
         for(k in (j+1):nobs[i]) 
          count_un[(ti[j,1]),(ti[k,1])] = count_un[(ti[j,1]),(ti[k,1])]+1       
    }
   }

  }





  


 np=dim(x)[2]+1 #Number of regression np (betas)



#Getting initial parameter estimates assuming independence

Data_Set=cbind(id,y,x)

dataframe=as.data.frame(Data_Set)

if (as.matrix(dim(x))[2,1] == 1 ) hi=glm(V2~V3,family = fam, data=dataframe)
if (as.matrix(dim(x))[2,1] == 2 ) hi=glm(V2~V3 + V4,family = fam, data=dataframe)
if (as.matrix(dim(x))[2,1] == 3 ) hi=glm(V2~V3 + V4 + V5,family = fam, data=dataframe)


Beta=hi$coefficients
Beta=as.matrix(Beta)
beta=matrix(0,np,1)
for(i in 1:np)
 beta[i,1]=Beta[i,1]










 betadiff <- 1
 iteration <- 0
 betanew <- beta


 N=sum(nobs)  #Used in estimating the scale parameter.


 one=matrix(1,length(id),1)
 x=cbind(one,x)              #Design matrix





 while (betadiff > tol && iteration < maxiter)   
 {
  beta <- betanew
  GEE=matrix(0,np,1)  #This will be the summation of the n components of the GEE. 
  I=matrix(0,np,np)    #This will be the sum of the components in the variance equation

  meat=0   #This is the "meat" of the empirical variance estimate
 

  I_ind=matrix(0,np,np)    #Used for CIC using model-based independence


   #Now for obtaining the sums (from each independent subject or cluster) used in estimating the correlation parameter(s) and also scale parameter.  

   if ((corstr == "exchangeable") | (corstr == "AR-1") | (corstr == "MA-1")) sum = 0

   if (corstr == "Toeplitz") sum = matrix(0,(max_time-1),1)

   if (corstr == "unstructured")  sum = diag(max_time) 


   sum_scale=0



  loc1 <- 0
  loc2 <- 0
  for(i in 1:nsub){
   loc1 <- loc2 + 1
   loc2 <- loc1 + nobs[i] - 1
   yi <- as.matrix(y[loc1:loc2, ])
   xi <- x[loc1:loc2, ]
   ni <- nrow(yi)
   ti <- as.matrix(time[loc1:loc2, ])


   if (fam == "gaussian") {     
                ui <- xi %*% beta
                fui <- ui
                fui_dev <- diag(ni)
                vui <- diag(ni)
                Pearson = (yi-ui)     
   }
   if (fam == "poisson") {
                ui <- exp(xi %*% beta)
                fui <- log(ui)
                fui_dev <- diag(as.vector(ui))
                vui <- diag(as.vector(sqrt(1/ui)))
                Pearson = (yi-ui)/sqrt(ui)
   }
   if (fam == "gamma") {
                ui <- 1/(xi %*% beta)
                fui <- 1/ui
                fui_dev <- -diag(as.vector(ui)) %*% diag(as.vector(ui))
                vui <- diag(as.vector(1/ui))
                Pearson = (yi-ui)/ui
   }
   if (fam == "binomial") {
                ui <- 1/(1 + exp(-xi %*% beta))
                fui <- log(ui) - log(1 - ui)
                fui_dev <- diag(as.vector(ui)) %*% diag(as.vector(1 - ui))
                vui <- diag(as.vector(sqrt(1/ui))) %*% diag(as.vector(sqrt(1/(1 - ui))))
                Pearson = (yi-ui)/sqrt(ui*(1-ui))
   }


   
   for(j in 1:ni)
    sum_scale = sum_scale + Pearson[j,1]^2
 

   if (corstr == "exchangeable") {
   	for(j in 1:(ni-1))
         for(k in (j+1):ni)
          sum = sum + Pearson[j,1]*Pearson[k,1]        	
   }


   if (corstr == "AR-1" | corstr == "MA-1") {
    if(nobs[i]>1)
    {
     for(l in 1:(nobs[i]-1))
      if( (ti[(l+1),1]-ti[l,1]) == 1)
       sum = sum + Pearson[l,1]*Pearson[(l+1),1] 
    }
   }


   if (corstr == "Toeplitz") {
    if(nobs[i]>1)
    {
     for(k in 1:(max_time-1))
      for(l in 1:(nobs[i]-1))
       for(j in (l+1):nobs[i])
        if( (ti[j,1]-ti[l,1]) == k)
         sum[k,1] = sum[k,1] + Pearson[j,1]*Pearson[l,1] 
    }
   }


   if (corstr == "unstructured") {
    if(nobs[i]>1)
    {
	for(j in 1:(nobs[i]-1))
         for(k in (j+1):nobs[i])
          {
           sum[(ti[j,1]),(ti[k,1])] = sum[(ti[j,1]),(ti[k,1])] + Pearson[j,1]*Pearson[k,1]
           sum[(ti[k,1]),(ti[j,1])] = sum[(ti[k,1]),(ti[j,1])] + Pearson[j,1]*Pearson[k,1]
          }            
    }
   } 


  } #end of the for loop (completing adding up the GEE, I, and residual terms for each independent subject) 





  scale = sum_scale/(sum(nobs)-np)

  if (corstr == "exchangeable") corr_par = sum/(N_star-np)/scale
  
  if (corstr == "AR-1" | corstr == "MA-1") corr_par = sum/(K_1-np)/scale

  if (corstr == "Toeplitz") 
  {
   corr_par_matrix=matrix(0,(max_time-1),1)
   for(k in 1:(max_time-1))
    corr_par_matrix[k,1] = sum[k,1]/(Toep_corr[k,1]-np)/scale
  }




   if (corstr == "unstructured") {
 	       for(j in 1:(max_time-1))
                for(k in (j+1):max_time) 
                {              
                 sum[j,k] = sum[j,k]/(count_un[j,k]-np)/scale
                 sum[k,j] = sum[k,j]/(count_un[j,k]-np)/scale
                }              
               Ri_max = sum
   }

  


                

  loc1 <- 0
  loc2 <- 0
  for(i in 1:nsub){
   loc1 <- loc2 + 1
   loc2 <- loc1 + nobs[i] - 1
   yi <- as.matrix(y[loc1:loc2, ])
   xi <- x[loc1:loc2, ]
   ni <- nrow(yi)
   ti <- as.matrix(time[loc1:loc2, ])


   if (fam == "gaussian") {            
                ui <- xi %*% beta
                fui <- ui
                fui_dev <- diag(ni)
                vui <- diag(ni) 
   }
   if (fam == "poisson") {
                ui <- exp(xi %*% beta)
                fui <- log(ui)
                fui_dev <- diag(as.vector(ui))
                vui <- diag(as.vector(sqrt(1/ui)))
   }
   if (fam == "gamma") {
                ui <- 1/(xi %*% beta)
                fui <- 1/ui
                fui_dev <- -diag(as.vector(ui)) %*% diag(as.vector(ui))
                vui <- diag(as.vector(1/ui))
   }
   if (fam == "binomial") {
                ui <- 1/(1 + exp(-xi %*% beta))
                fui <- log(ui) - log(1 - ui)
                fui_dev <- diag(as.vector(ui)) %*% diag(as.vector(1 - ui))
                vui <- diag(as.vector(sqrt(1/ui))) %*% diag(as.vector(sqrt(1/(1 - ui))))
   }




   #Obtaining the sums used in the GEE for the betas

   if (corstr == "independence" ) {
                wi <- t(xi) %*% fui_dev %*% vui %*% vui
                GEE <- GEE + wi %*% (yi - ui)
                I = I + wi %*% fui_dev %*% xi
   }


   if (corstr == "unstructured") {
	       Ri_i=diag(ni)
               for(k in 1:(ni-1))
                for(l in (k+1):ni)                
                 Ri_i[k,l]=Ri_i[l,k]=Ri_max[(ti[k,1]),(ti[l,1])]

               Ri_inv = solve(Ri_i)         
               wi <- t(xi) %*% fui_dev %*% vui %*% Ri_inv %*% vui
               GEE <- GEE + wi %*% (yi - ui)
               I = I + wi %*% fui_dev %*% xi
   }

   if (corstr == "exchangeable") {
               Ri = diag(ni) + matrix(corr_par,ni,ni) - diag(corr_par,ni)
               Ri_inv = solve(Ri)           
               wi <- t(xi) %*% fui_dev %*% vui %*% Ri_inv %*% vui
               GEE <- GEE + wi %*% (yi - ui)
               I = I + wi %*% fui_dev %*% xi
   }
   

   if (corstr == "AR-1") {
               Ri = matrix(0,ni,ni)                       
               for(k in 1:ni)
                for(l in 1:ni)
                 Ri[k,l]=corr_par^abs(ti[k,1]-ti[l,1]) 

               Ri_inv = solve(Ri)           
               wi <- t(xi) %*% fui_dev %*% vui %*% Ri_inv %*% vui
               GEE <- GEE + wi %*% (yi - ui)
               I = I + wi %*% fui_dev %*% xi
   }
   

   if (corstr == "MA-1") {
               Ri = diag(ni)                        
               for(k in 1:(ni-1))
                if( (ti[(k+1),1]-ti[k,1])==1 )
                 Ri[k,(k+1)]=Ri[(k+1),k]=corr_par

               Ri_inv = solve(Ri)           
               wi <- t(xi) %*% fui_dev %*% vui %*% Ri_inv %*% vui
               GEE <- GEE + wi %*% (yi - ui)
               I = I + wi %*% fui_dev %*% xi
   }
   

   if (corstr == "Toeplitz") {
               Ri = diag(ni)                        
               for(j in 1:(ni-1))             
                for(k in (j+1):ni)
                 {
                  position = (ti[k,1]-ti[j,1])
                  Ri[k,j]=Ri[j,k]=corr_par_matrix[position,1]
                 }

               Ri_inv = solve(Ri)           
               wi <- t(xi) %*% fui_dev %*% vui %*% Ri_inv %*% vui
               GEE <- GEE + wi %*% (yi - ui)
               I = I + wi %*% fui_dev %*% xi
   }


   I_ind = I_ind + t(xi) %*% fui_dev %*%  vui %*% vui %*% fui_dev %*% xi  

   meat = meat + wi %*% (yi - ui) %*% t((yi - ui)) %*% t(wi)   


  } #end of the for loop

 
  betanew <- beta + solve(I) %*% GEE                             #Newton Raphson 
                                                                 
  betadiff <- sum(abs(betanew - beta))                           #Determining if convergence or not


  
  #Now getting the model-based and empirical covariance matrices
  
  emp = solve(I) %*% meat %*% solve(I)

  Model_basedVar = scale*solve(I)


  iteration <- iteration + 1                                     #Counting how many iterations it has gone through
  #print(iteration)
  #print(beta)
  #print(corr_par)
  #print(scale)
 } #End of the while loop

 
 non_converge=0
 if(iteration==maxiter) non_converge=1 



 if (fam != "binomial") 
  I_ind = I_ind/scale








G=matrix(0,np,np)

if(corstr != "independence")
{



 #Obtaining the covariance inflations 
 #Obtaining G to correct for the source of bias in the estimated asymptotic covariance matrix 


   if ((corstr == "exchangeable") | (corstr == "AR-1") | (corstr == "MA-1")) sum_deriv = matrix(0,np,1)  

   if (corstr == "Toeplitz") sum_deriv = matrix(0,(max_time-1),np) #

   if (corstr == "unstructured") sum_deriv = matrix(0,max_time,(max_time*np))


  loc1 <- 0
  loc2 <- 0
  for(i in 1:nsub){
   loc1 <- loc2 + 1
   loc2 <- loc1 + nobs[i] - 1
   yi <- as.matrix(y[loc1:loc2, ])
   xi <- x[loc1:loc2, ]
   ni <- nrow(yi)
   ti <- as.matrix(time[loc1:loc2, ])


   if (fam == "gaussian") {           
                ui <- xi %*% beta
                fui <- ui
                fui_dev <- diag(ni)
                vui <- diag(ni)
                Pearson = (yi-ui)

                Di = fui_dev %*% xi 
   }
   if (fam == "poisson") {
                ui <- exp(xi %*% beta)
                fui <- log(ui)
                fui_dev <- diag(as.vector(ui))
                fui_dev2 <- diag(as.vector(sqrt(ui)))  #Since we take the derivative of the Pearson Residual, not just the residual
                vui <- diag(as.vector(sqrt(1/ui)))
                Pearson = (yi-ui)/sqrt(ui)

                Di = fui_dev2 %*% xi 
   }
   if (fam == "gamma") {
                ui <- 1/(xi %*% beta)
                fui <- 1/ui
                fui_dev <- -diag(as.vector(ui)) %*% diag(as.vector(ui))
                fui_dev2 <- -diag(as.vector(matrix(1,ni,ni))) %*% diag(as.vector(matrix(1,ni,ni)))  #Since we take the derivative of the Pearson Residual, not just the residual
                vui <- diag(as.vector(1/ui))
                Pearson = (yi-ui)/ui

                Di = fui_dev2 %*% xi 
   }
   if (fam == "binomial") {
                ui <- 1/(1 + exp(-xi %*% beta))
                fui <- log(ui) - log(1 - ui)
                fui_dev <- diag(as.vector(ui)) %*% diag(as.vector(1 - ui))
                fui_dev2 <- diag(as.vector(sqrt(ui))) %*% diag(as.vector(sqrt(1 - ui)))    #Since we take the derivative of the Pearson Residual, not just the residual
                vui <- diag(as.vector(sqrt(1/ui))) %*% diag(as.vector(sqrt(1/(1 - ui))))
                Pearson = (yi-ui)/sqrt(ui*(1-ui))

                Di = fui_dev2 %*% xi 
   }


   ei <- Pearson 


   if (corstr == "exchangeable") {
       for(l in 1:np)
   	for(j in 1:(ni-1))
         for(k in (j+1):ni)
          sum_deriv[l,1] = sum_deriv[l,1] - Di[j,l]*ei[k,1] - Di[k,l]*ei[j,1]      	
   }


   if (corstr == "AR-1" | corstr == "MA-1") {
    if(nobs[i]>1)
    {
       for(l in 1:np)
	for(j in 1:(ni-1))
         if( (ti[(j+1),1]-ti[j,1]) == 1)
          sum_deriv[l,1] = sum_deriv[l,1] - Di[j,l]*ei[(j+1),1] - Di[(j+1),l]*ei[j,1]                                                     
    }                                                     
   }


   if (corstr == "Toeplitz") {
    if(nobs[i]>1)
    {
    for(l in 1:np)
     for(k in 1:(max_time-1))
      for(f in 1:(nobs[i]-1))
       for(j in (f+1):nobs[i])
        if( (ti[j,1]-ti[f,1]) == k)
         sum_deriv[k,l] = sum_deriv[k,l] - Di[j,l]*ei[f,1] - Di[f,l]*ei[j,1]
    }
   }



   if (corstr == "unstructured") 
   {
    location1=0
    location2=0
    for(l in 1:np)
    {
     location1 <- location2 + 1
     location2 <- location1 + max_time - 1
     temp=matrix(0,max_time,max_time)

     if(nobs[i]>1)
     {
	for(j in 1:(nobs[i]-1))
         for(k in (j+1):nobs[i])
          {
           temp[(ti[j,1]),(ti[k,1])] = temp[(ti[j,1]),(ti[k,1])] - Di[j,l]*ei[k,1] - Di[k,l]*ei[j,1]
           temp[(ti[k,1]),(ti[j,1])] = temp[(ti[k,1]),(ti[j,1])] - Di[j,l]*ei[k,1] - Di[k,l]*ei[j,1]
          }            
     }    

     sum_deriv[,location1:location2] = sum_deriv[,location1:location2] + temp
    }
   }




  } #end of the for loop 











 G=matrix(0,np,np)

  loc1 <- 0
  loc2 <- 0
  for(i in 1:nsub){
   loc1 <- loc2 + 1
   loc2 <- loc1 + nobs[i] - 1
   yi <- as.matrix(y[loc1:loc2, ])
   xi <- x[loc1:loc2, ]
   ni <- nrow(yi)
   ti <- as.matrix(time[loc1:loc2, ])


   if (fam == "gaussian") {           
                ui <- xi %*% beta
                fui <- ui
                fui_dev <- diag(ni)
                vui <- diag(ni)   
   }
   if (fam == "poisson") {
                ui <- exp(xi %*% beta)
                fui <- log(ui)
                fui_dev <- diag(as.vector(ui))
                vui <- diag(as.vector(sqrt(1/ui)))
   }
   if (fam == "gamma") {
                ui <- 1/(xi %*% beta)
                fui <- 1/ui
                fui_dev <- -diag(as.vector(ui)) %*% diag(as.vector(ui))
                vui <- diag(as.vector(1/ui))
   }
   if (fam == "binomial") {
                ui <- 1/(1 + exp(-xi %*% beta))
                fui <- log(ui) - log(1 - ui)
                fui_dev <- diag(as.vector(ui)) %*% diag(as.vector(1 - ui))
                vui <- diag(as.vector(sqrt(1/ui))) %*% diag(as.vector(sqrt(1/(1 - ui))))
   }



   if (corstr == "exchangeable") {
               Ri = diag(ni) + matrix(corr_par,ni,ni) - diag(corr_par,ni)   
               Ri_inv = solve(Ri)           
   }
   

   if (corstr == "AR-1") {
               Ri = matrix(0,ni,ni)                     
               for(k in 1:ni)
                for(l in 1:ni)
                 Ri[k,l]=corr_par^abs(ti[k,1]-ti[l,1]) 

               Ri_inv = solve(Ri)             
   }
   

   if (corstr == "MA-1") {
               Ri = diag(ni)                      
               for(k in 1:(ni-1))
                if( (ti[(k+1),1]-ti[k,1])==1 )
                 Ri[k,(k+1)]=Ri[(k+1),k]=corr_par

               Ri_inv = solve(Ri)           
   }
  

   if (corstr == "Toeplitz") {
               Ri = diag(ni)                      
               for(j in 1:(ni-1))             
                for(k in (j+1):ni)
                 {
                  position = (ti[k,1]-ti[j,1])
                  Ri[k,j]=Ri[j,k]=corr_par_matrix[position,1]
                 }

               Ri_inv = solve(Ri)           
   }




   if (corstr == "unstructured") {
	       Ri_i=diag(ni)
               for(k in 1:(ni-1))
                for(l in (k+1):ni)                
                 Ri_i[k,l]=Ri_i[l,k]=Ri_max[(ti[k,1]),(ti[l,1])]

               Ri_inv = solve(Ri_i)         
   }












   location1=0
   location2=0

   for(l in 1:np)
   {

    if (corstr == "exchangeable")
    {
     element = sum_deriv[l,1]/(N_star-np)/scale
     Ri_deriv = matrix(element,ni,ni) - diag(element,ni)
    }  

    if (corstr == "MA-1")
    {
     element = sum_deriv[l,1]/(K_1-np)/scale
     Ri_deriv = matrix(0,ni,ni)                         
     for(k in 1:(ni-1))
      if( (ti[(k+1),1]-ti[k,1])==1 )
       Ri_deriv[k,(k+1)]=Ri_deriv[(k+1),k]=element
    } 

    if (corstr == "AR-1")
    {
     element = sum_deriv[l,1]/(K_1-np)/scale
     Ri_deriv = matrix(0,ni,ni)
     for(k in 1:ni)
      for(j in 1:ni)
       Ri_deriv[k,j]=abs(ti[k,1]-ti[j,1])*(corr_par^(abs(ti[k,1]-ti[j,1])-1))*element  
    }


    if (corstr == "Toeplitz")
    {
     Ri_deriv = matrix(0,ni,ni)
               for(j in 1:(ni-1))             
                for(k in (j+1):ni)
                 {
                  position = (ti[k,1]-ti[j,1])
                  element = sum_deriv[position,l]/(Toep_corr[position,1]-np)/scale
                  Ri_deriv[k,j]=Ri_deriv[j,k]=element
                 }          
    }


    if (corstr == "unstructured") 
    {
     location1 <- location2 + 1
     location2 <- location1 + max_time - 1
     Ri_deriv_max = sum_deriv[,location1:location2]
 	       for(j in 1:(max_time-1))
                for(k in (j+1):max_time) 
                {              
                 Ri_deriv_max[j,k] = Ri_deriv_max[j,k]/(count_un[j,k]-np)/scale
                 Ri_deriv_max[k,j] = Ri_deriv_max[k,j]/(count_un[j,k]-np)/scale
                }   

	       Ri_deriv=matrix(0,ni,ni)
               for(k in 1:(ni-1))
                for(j in (k+1):ni)                
                 Ri_deriv[k,j]=Ri_deriv[j,k]=Ri_deriv_max[(ti[k,1]),(ti[j,1])]
    } 



    G[,l] = G[,l] - solve(I) %*% t(xi) %*% fui_dev %*% vui %*% Ri_inv %*% Ri_deriv %*% Ri_inv %*% vui %*% (yi-ui) 
   }
         
  } #end of the for loop that sums through all indpendent clusters/subjects 





} #end of the if(corstr =! "independence") part






















  #Getting the model-based covariance matrix with correction

  MB_correction = (diag(np)+G) %*% Model_basedVar %*% t(diag(np)+G)


 
  #Now getting the bias-corrected covariance matrix with and without (diag(np)+G) correction

  meat_BCMD=0              #This is the "meat" of the bias-corrected empirical variance estimate from Mancl & DeRouen
  meat_BCKC=0              #This is the "meat" of the bias-corrected empirical variance estimate from Kauermann & Carroll
  meat_BCFG=0              #This is the "meat" of the bias-corrected empirical variance estimate from Fay & Graubard
  meat_BCFZZ=0             #This is the "meat" of the bias-corrected empirical variance estimate from FZZ
  meat_BCGSTpt1=matrix(0,max_time,max_time)
  meat_BCGST=0
  
  meat_BCMBN=0
  meat_FZZ=0
  
  meat_BCMD_correction=0   #This is the "meat" of the bias-corrected empirical variance estimate from Mancl & DeRouen
  meat_BCKC_correction=0   #This is the "meat" of the bias-corrected empirical variance estimate from Kauermann & Carroll
  meat_BCFG_correction=0   #This is the "meat" of the bias-corrected empirical variance estimate from Fay & Graubard
  meat_BCFZZ_correction=0   #This is the "meat" of the bias-corrected empirical variance estimate from FZZ
  
  meat_emp_correction=0  

  #Also getting the unstructured components of the meat (methods of Pan (2001) and Long and Wang (2011) )
  R_Unst_typical = matrix(0,max_time,max_time)  #Obtaining the unstructured corr matrix that Pan (2001) proposed in the meat of the sandwich/empirical covariance matrix
  R_Unst_MD = matrix(0,max_time,max_time)       #Obtaining the unstructured corr matrix that Long and Wang (2001) proposed in the meat of the sandwich/empirical covariance matrix, using the Mancl & DeRouen correction
  counting = matrix(0,max_time,max_time)        #Counting the number of summations inside the unstructured correlation                       


  #Working on getting the df based on the Pan and Wall (2002) approach
   P_matrixMD = P_matrixKC = P_matrixFG  = P_matrixFZZ = P_matrixGSH = P_matrixMBN = NULL


  #######For new df based on Pan and Wall (2001) formulas
  #Obtaining the K_i's
  K_MD = K_KC = K_FG = K_FZZ = K_GSH = K_MBN = NULL

  loc1 <- 0
  loc2 <- 0
  for(i in 1:nsub){
   loc1 <- loc2 + 1
   loc2 <- loc1 + nobs[i] - 1
   yi <- as.matrix(y[loc1:loc2, ])
   xi <- x[loc1:loc2, ]
   ni <- nrow(yi)
   ti <- as.matrix(time[loc1:loc2, ])


   if (fam == "gaussian") {            
                ui <- xi %*% beta
                fui <- ui
                fui_dev <- diag(ni)
                vui <- diag(ni)
                Pearson = (yi-ui)     
   }
   if (fam == "poisson") {
                ui <- exp(xi %*% beta)
                fui <- log(ui)
                fui_dev <- diag(as.vector(ui))
                vui <- diag(as.vector(sqrt(1/ui)))
                Pearson = (yi-ui)/sqrt(ui)
   }
   if (fam == "gamma") {
                ui <- 1/(xi %*% beta)
                fui <- 1/ui
                fui_dev <- -diag(as.vector(ui)) %*% diag(as.vector(ui))
                vui <- diag(as.vector(1/ui))
                Pearson = (yi-ui)/ui
   }
   if (fam == "binomial") {
                ui <- 1/(1 + exp(-xi %*% beta))
                fui <- log(ui) - log(1 - ui)
                fui_dev <- diag(as.vector(ui)) %*% diag(as.vector(1 - ui))
                vui <- diag(as.vector(sqrt(1/ui))) %*% diag(as.vector(sqrt(1/(1 - ui))))
                Pearson = (yi-ui)/sqrt(ui*(1-ui))
   }






   if (corstr == "independence" ) {
                wi <- t(xi) %*% fui_dev %*% vui %*% vui
                Di = fui_dev %*% xi
   }


   if (corstr == "unstructured") {
	       Ri_i=diag(ni)
               for(k in 1:(ni-1))
                for(l in (k+1):ni)                
                 Ri_i[k,l]=Ri_i[l,k]=Ri_max[(ti[k,1]),(ti[l,1])]

               Ri_inv = solve(Ri_i)         
               wi <- t(xi) %*% fui_dev %*% vui %*% Ri_inv %*% vui
               Di = fui_dev %*% xi
   }


   if (corstr == "exchangeable") {
               Ri = diag(ni) + matrix(corr_par,ni,ni) - diag(corr_par,ni)     
               Ri_inv = solve(Ri)           
               wi <- t(xi) %*% fui_dev %*% vui %*% Ri_inv %*% vui
               Di = fui_dev %*% xi
   }
   

   if (corstr == "AR-1") {
               Ri = matrix(0,ni,ni)                       
               for(k in 1:ni)
                for(l in 1:ni)
                 Ri[k,l]=corr_par^abs(ti[k,1]-ti[l,1])  

               Ri_inv = solve(Ri)           
               wi <- t(xi) %*% fui_dev %*% vui %*% Ri_inv %*% vui
               Di = fui_dev %*% xi
   }
   

   if (corstr == "MA-1") {
               Ri = diag(ni)                        
               for(k in 1:(ni-1))
                if( (ti[(k+1),1]-ti[k,1])==1 )
                 Ri[k,(k+1)]=Ri[(k+1),k]=corr_par

               Ri_inv = solve(Ri)           
               wi <- t(xi) %*% fui_dev %*% vui %*% Ri_inv %*% vui
               Di = fui_dev %*% xi
   }
   

   if (corstr == "Toeplitz") {
               Ri = diag(ni)                        
               for(j in 1:(ni-1))             
                for(k in (j+1):ni)
                 {
                  position = (ti[k,1]-ti[j,1])
                  Ri[k,j]=Ri[j,k]=corr_par_matrix[position,1]
                 }

               Ri_inv = solve(Ri)           
               wi <- t(xi) %*% fui_dev %*% vui %*% Ri_inv %*% vui
               Di = fui_dev %*% xi
   }


  

   H = Di%*%solve(I)%*%wi 
   meat_BCMD = meat_BCMD + wi %*% solve(diag(ni)-H) %*% (yi - ui) %*% t((yi - ui)) %*% solve(diag(ni)-t(H)) %*% t(wi)
   meat_BCKC = meat_BCKC + wi %*% solve(diag(ni)-H) %*% (yi - ui) %*% t((yi - ui)) %*% t(wi)

      
   if (fam == "gaussian") 
     Ai = diag(ni)
   
   if (fam == "poisson") 
     Ai = diag(as.vector(ui))
   
   
   if (fam == "gamma") 
     Ai = diag(as.vector(ui)) %*% diag(as.vector(ui))
   
   if (fam == "binomial") 
     Ai = diag(as.vector(ui)) %*% diag(as.vector(1 - ui))   
   
  ind_GST_SUM=matrix(0,max_time,max_time)
  ind_GST_SUM[1:ni,1:ni] = solve(sqrt(Ai)) %*% (yi - ui) %*% t((yi - ui)) %*% solve(sqrt(Ai))
   meat_BCGSTpt1 = meat_BCGSTpt1 +   ind_GST_SUM
   
   if ( nsub >3*np ) delMBN = np/(nsub-np) 
   if ( nsub <= 3*np) delMBN=0.5
   KMBN = (nsub*ni-1)*nsub / ((nsub-1) * (nsub*ni-np))
   #epsilon is max 1 or trace of model based var * meat of LZ
   epsMBN = max(1, matrix.trace(Model_basedVar%*%meat)/np) # 1 is default value of r.
   
   meat_MBN_i = wi %*% ((KMBN * (yi - ui) %*% t((yi - ui)))) %*% t(wi) + epsMBN*delMBN*Model_basedVar
   
   meat_BCMBN = meat_BCMBN + meat_MBN_i
   
   #meat_FZZ is similar to meat_BCMD but contains a correction factor; calculating that here.
   
   for(j in 1:nsub)     
   {
     
     loc1_forFZZ <- 0
     loc2_forFZZ <- 0
     for(i in 1:nsub){
       loc1_forFZZ <- loc1
       loc2_forFZZ <- loc2
       yj <- yi
       xj <- xi
       nj <- ni
       tj <- as.matrix(time[loc1_forFZZ:loc2_forFZZ, ])
       
       
       SUM=0
       if((j !=i))
       {
         if (fam == "gaussian") {         
           uj <- xj %*% beta
           residualj= (yj-uj)
           vuj = diag(nj)
           fuj <- uj
           fuj_dev <- diag(nj)
         }
         if (fam == "poisson") {
           uj <- exp(xj %*% beta)
           fuj <- log(uj)
           fuj_dev <- diag(as.vector(uj))
           vuj <- diag(as.vector(sqrt(1/uj)))
           residualj = (yj-uj)/sqrt(uj)
         }
         if (fam == "gamma") {
           uj <- 1/(xj %*% beta)
           fui <- 1/uj
           fuj_dev <- -diag(as.vector(uj)) %*% diag(as.vector(uj))
           vuj <- diag(as.vector(1/uj))
           residualj = (yj-uj)/uj
         }
         if (fam == "binomial") {
           uj <- 1/(1 + exp(-xj %*% beta))
           fuj <- log(uj) - log(1 - uj)
           fuj_dev <- diag(as.vector(uj)) %*% diag(as.vector(1 - uj))
           vuj <- diag(as.vector(sqrt(1/uj))) %*% diag(as.vector(sqrt(1/(1 - uj))))
           residualj = (yj-uj)/sqrt(uj*(1-uj))
         }
         
         
         
         if (corstr == "independence" ) {
           wj <- t(xj) %*% fuj_dev %*% vuj %*% vuj
           Dj = fuj_dev %*% xj
         }
         
         
         if (corstr == "exchangeable") {
           Rj = diag(nj) + matrix(corr_par,nj,nj) - diag(corr_par,nj)     
           Rj_inv = solve(Rj)           
           wj <- t(xj) %*% fuj_dev %*% vuj %*% Rj_inv %*% vuj
           Dj = fuj_dev %*% xj
         }
         
         if (corstr == "AR-1") {
           Rj = matrix(0,nj,nj)                       
           for(k in 1:nj)
             for(l in 1:nj)
               Rj[k,l]=corr_par^abs(k-l)  
             
             Rj_inv = solve(Rj)           
             wj <- t(xj) %*% fuj_dev %*% vuj %*% Rj_inv %*% vuj
             Dj = fuj_dev %*% xj
         }
         
         
         
         Hj = Dj %*% solve(I) %*% wj 
         SUM = SUM + (Hj^2)%*%(residualj^2)  
       }
     }
   } #Ends calculations for meat FZZ part
   
   meat_FZZ = meat_FZZ + wi %*% solve(diag(ni)-H) %*% (yi - ui - SUM) %*% t((yi - ui - SUM)) %*% solve(diag(ni)-t(H)) %*% t(wi)

   #Calculating Fay and Graubard (2001)
   H_Z =t(Di)%*%t(wi)%*%solve(I)
   Z=matrix(0,np,np)
   for(i in 1:np) 
   {
     if (H_Z[i,i]<=0.75) Z[i,i]=(1-H_Z[i,i])^(-1/2)
     if (H_Z[i,i]>0.75) Z[i,i]=(1-0.75)^(-1/2)
   }
   meat_BCFG = meat_BCFG + Z %*% wi %*% (yi - ui) %*% t((yi - ui)) %*% t(wi) %*% Z
   
   
   H_c = Di%*%(diag(np)+G)%*%solve(I)%*%wi
   meat_BCMD_correction = meat_BCMD_correction + wi %*% solve(diag(ni)-H_c) %*% (yi - ui) %*% t((yi - ui)) %*% solve(diag(ni)-t(H_c)) %*% t(wi)
   meat_BCKC_correction = meat_BCKC_correction + wi %*% solve(diag(ni)-H_c) %*% (yi - ui) %*% t((yi - ui)) %*% t(wi)
   meat_FZZ_correction = meat_BCMD_correction + wi %*% solve(diag(ni)-H_c) %*% (yi - ui - SUM) %*% t((yi - ui - SUM)) %*% solve(diag(ni)-t(H_c)) %*% t(wi)
   meat_emp_correction = meat_emp_correction + wi %*% (yi - ui) %*% t((yi - ui)) %*% t(wi)

   H_Z_c =(diag(np)+G)%*%t(Di)%*%t(wi)%*%solve(I)
   #print(G)
   #print(solve(I))
   #print(t(Di))
   Z_c=matrix(0,np,np)
   for(i in 1:np)
   {
     if (H_Z_c[i,i]<=0.75) Z_c[i,i]=(1-H_Z_c[i,i])^(-1/2)
     if (H_Z_c[i,i]>0.75) Z_c[i,i]=(1-0.75)^(-1/2)
   }
   meat_BCFG_correction = meat_BCFG_correction + Z_c %*% wi %*% (yi - ui) %*% t((yi - ui)) %*% t(wi) %*% Z_c   
   

   #Obtaining the matrices needed to calculate the Pan and Wall df
   P_MD_i = P_KC_i = P_FG_i = P_AVG_i = P_FZZ_i = P_WL_i = P_GSH_i = P_MBN_i = NULL
   meat_MD_i = wi %*% solve(diag(ni)-H_c) %*% (yi - ui) %*% t((yi - ui)) %*% solve(diag(ni)-t(H_c)) %*% t(wi)
   meat_KC_i = wi %*% solve(diag(ni)-H_c) %*% (yi - ui) %*% t((yi - ui)) %*% t(wi)
   meat_FG_i = Z %*% wi %*% (yi - ui) %*% t((yi - ui)) %*% t(wi) %*% Z
   meat_FZZ_i = wi %*% solve(diag(ni)-H_c) %*% (yi - ui - SUM) %*% t((yi - ui - SUM)) %*% solve(diag(ni)-t(H_c)) %*% t(wi)
   
 
   for(j in 1:np)
   {
     P_MD_i=c(P_MD_i,meat_MD_i[,j])
     P_KC_i=c(P_KC_i,meat_KC_i[,j])
     P_FG_i=c(P_FG_i,meat_FG_i[,j])
     P_FZZ_i=c(P_FZZ_i,meat_FZZ_i[,j])
     P_MBN_i=c(P_MBN_i,meat_MBN_i[,j])
   }
   
   P_MD_i=as.matrix(P_MD_i)
   P_KC_i=as.matrix(P_KC_i)
   P_FG_i=as.matrix(P_FG_i)
   P_FZZ_i=as.matrix(P_FZZ_i)
   P_MBN_i=as.matrix(P_MBN_i)
   
   
   P_matrixMD = cbind(P_matrixMD,P_MD_i)
   P_matrixKC = cbind(P_matrixKC,P_KC_i)
   P_matrixFG = cbind(P_matrixKC,P_FG_i)   
   P_matrixFZZ = cbind(P_matrixFZZ,P_FZZ_i)   
   P_matrixMBN = cbind(P_matrixMBN,P_MBN_i)
   


   #Getting the unstructured components of the meat (methods of Pan (2001) and Wang and Long (2011))   
    Pearson2 = solve(diag(ni)-H_c) %*% Pearson

 	       for(j in 1:ni)
                for(k in 1:ni) 
                {              
                 R_Unst_typical[(ti[j,1]),(ti[k,1])] = R_Unst_typical[(ti[j,1]),(ti[k,1])] + Pearson[j,1]*Pearson[k,1]
                 R_Unst_MD[(ti[j,1]),(ti[k,1])] = R_Unst_MD[(ti[j,1]),(ti[k,1])] + Pearson2[j,1]*Pearson2[k,1]
                 counting[(ti[j,1]),(ti[k,1])]=counting[(ti[j,1]),(ti[k,1])]+1
                }              
 

    ##################For new df based on Pan (2001) formulas
    k_MD_i = solve(diag(ni)-H_c) %*% (yi - ui) %*% t((yi - ui)) %*% solve(diag(ni)-t(H_c))
    k_KC_i = solve(diag(ni)-H_c) %*% (yi - ui) %*% t((yi - ui)) 
    k_FG_i = (yi - ui) %*% t((yi - ui)) 
    k_FZZ_i = solve(diag(ni)-H_c) %*% (yi - ui - SUM) %*% t((yi - ui - SUM)) %*% solve(diag(ni)-t(H_c))
    k_MBN_i = ((KMBN * (yi - ui) %*% t((yi - ui))))
    

    K_MD_i_init = K_KC_i_init = K_FG_i_init = K_FZZ_i_init = K_MBN_i_init = NULL
    K_MD_i = K_KC_i = K_FG_i = K_FZZ_i = K_MBN_i = matrix(0,max_time*max_time,1)
    for(j in 1:ni)
    {
      K_MD_i_init=c(K_MD_i_init,k_MD_i[,j])
      K_KC_i_init=c(K_KC_i_init,k_KC_i[,j])
      K_FG_i_init=c(K_FG_i_init,k_FG_i[,j])
      K_FZZ_i_init=c(K_FZZ_i_init,k_FZZ_i[,j])
      K_MBN_i_init=c(K_MBN_i_init,k_MBN_i[,j])     
    }
    K_MD_i[1:(ni*ni),1]=as.matrix(K_MD_i_init) 
    K_KC_i[1:(ni*ni),1]=as.matrix(K_KC_i_init)
    K_FG_i[1:(ni*ni),1]=as.matrix(K_FG_i_init)
    K_FZZ_i[1:(ni*ni),1]=as.matrix(K_FZZ_i_init)
    K_MBN_i[1:(ni*ni),1]=as.matrix(K_MBN_i_init)
    
    
    K_MD = cbind(K_MD,K_MD_i)
    K_KC = cbind(K_KC,K_KC_i)
    K_FG = cbind(K_FG,K_FG_i)   
    K_FZZ = cbind(K_FZZ,K_FZZ_i)
    K_MBN = cbind(K_MBN,K_MBN_i) 
       
  } #end of the for loop 

  #Have to calculate Gosho et al. separately as it has a unique sum; doing this in the next portion of code.
  P_matrixGSH = K_GSH =  NULL
  
  loc1 <- 0
  loc2 <- 0
  for(i in 1:nsub){
    loc1 <- loc2 + 1
    loc2 <- loc1 + nobs[i] - 1
    yi <- as.matrix(y[loc1:loc2, ])
    xi <- x[loc1:loc2, ]
    ni <- nrow(yi)
    ti <- as.matrix(time[loc1:loc2, ])
    
    
    if (fam == "gaussian") {      
      ui <- xi %*% beta
      fui <- ui
      fui_dev <- diag(ni)
      vui <- diag(ni)
      Pearson = (yi-ui)     
    }
    if (fam == "poisson") {
      ui <- exp(xi %*% beta)
      fui <- log(ui)
      fui_dev <- diag(as.vector(ui))
      vui <- diag(as.vector(sqrt(1/ui)))
      Pearson = (yi-ui)/sqrt(ui)
    }
    if (fam == "gamma") {
      ui <- 1/(xi %*% beta)
      fui <- 1/ui
      fui_dev <- -diag(as.vector(ui)) %*% diag(as.vector(ui))
      vui <- diag(as.vector(1/ui))
      Pearson = (yi-ui)/ui
    }
    if (fam == "binomial") {
      ui <- 1/(1 + exp(-xi %*% beta))
      fui <- log(ui) - log(1 - ui)
      fui_dev <- diag(as.vector(ui)) %*% diag(as.vector(1 - ui))
      vui <- diag(as.vector(sqrt(1/ui))) %*% diag(as.vector(sqrt(1/(1 - ui))))
      Pearson = (yi-ui)/sqrt(ui*(1-ui))
    }
    
    
    
    
    
    
    if (corstr == "independence" ) {
      wi <- t(xi) %*% fui_dev %*% vui %*% vui
      Di = fui_dev %*% xi
    }
    
    
    if (corstr == "unstructured") {
      Ri_i=diag(ni)
      for(k in 1:(ni-1))
        for(l in (k+1):ni)                
          Ri_i[k,l]=Ri_i[l,k]=Ri_max[(ti[k,1]),(ti[l,1])]
        
        Ri_inv = solve(Ri_i)         
        wi <- t(xi) %*% fui_dev %*% vui %*% Ri_inv %*% vui
        Di = fui_dev %*% xi
    }
    
    
    if (corstr == "exchangeable") {
      Ri = diag(ni) + matrix(corr_par,ni,ni) - diag(corr_par,ni)    
      Ri_inv = solve(Ri)           
      wi <- t(xi) %*% fui_dev %*% vui %*% Ri_inv %*% vui
      Di = fui_dev %*% xi
    }
    
    
    if (corstr == "AR-1") {
      Ri = matrix(0,ni,ni)                    
      for(k in 1:ni)
        for(l in 1:ni)
          Ri[k,l]=corr_par^abs(ti[k,1]-ti[l,1])  
        
        Ri_inv = solve(Ri)           
        wi <- t(xi) %*% fui_dev %*% vui %*% Ri_inv %*% vui
        Di = fui_dev %*% xi
    }
    
    
    if (corstr == "MA-1") {
      Ri = diag(ni)                       
      for(k in 1:(ni-1))
        if( (ti[(k+1),1]-ti[k,1])==1 )
          Ri[k,(k+1)]=Ri[(k+1),k]=corr_par
        
        Ri_inv = solve(Ri)           
        wi <- t(xi) %*% fui_dev %*% vui %*% Ri_inv %*% vui
        Di = fui_dev %*% xi
    }
    
    
    if (corstr == "Toeplitz") {
      Ri = diag(ni)                     
      for(j in 1:(ni-1))             
        for(k in (j+1):ni)
        {
          position = (ti[k,1]-ti[j,1])
          Ri[k,j]=Ri[j,k]=corr_par_matrix[position,1]
        }
      
      Ri_inv = solve(Ri)           
      wi <- t(xi) %*% fui_dev %*% vui %*% Ri_inv %*% vui
      Di = fui_dev %*% xi
    }
    
    

    if (fam == "gaussian") 
      Ai = diag(ni)
    
    if (fam == "poisson") 
      Ai = diag(as.vector(ui))
    
    
    if (fam == "gamma") 
      Ai = diag(as.vector(ui)) %*% diag(as.vector(ui))
    
    if (fam == "binomial") 
      Ai = diag(as.vector(ui)) %*% diag(as.vector(1 - ui))   
    
    
    meat_BCGST = meat_BCGST +  1/(nsub-np) *wi %*% sqrt(Ai) %*% meat_BCGSTpt1[1:ni,1:ni] %*% sqrt(Ai) %*% t(wi)
    
    meat_GSH_i = 1/(nsub-np) * wi %*% sqrt(Ai) %*% meat_BCGSTpt1[1:ni,1:ni] %*% sqrt(Ai) %*% t(wi)
    
    P_GSH_i =  NULL
    for(j in 1:np)
    {
      P_GSH_i=c(P_GSH_i,meat_GSH_i[,j])
    }
    
    P_GSH_i=as.matrix(P_GSH_i)
    
    P_matrixGSH = cbind(P_matrixGSH,P_GSH_i)

    k_GSH_i = 1/(nsub-np) * sqrt(Ai) %*% meat_BCGSTpt1[1:ni,1:ni] %*% sqrt(Ai)

  
    K_GSH_i_init =  NULL
    K_GSH_i = matrix(0,max_time*max_time,1)
    for(j in 1:ni)
    {
      K_GSH_i_init=c(K_GSH_i_init,k_GSH_i[,j])
    }
        K_GSH_i[1:(ni*ni),1]=as.matrix(K_GSH_i_init)

    K_GSH = cbind(K_GSH,K_GSH_i)

  } #end of the for loop 



  ########################For new df based on Pan (2001) formulas 
  #Obtaining the K_bars (sample means) and sample covariance matrix (estimate for Omega in center of main covariance estimate that is needed)
  K_barMD = as.matrix(apply(K_MD,1,mean))
  K_barKC = as.matrix(apply(K_KC,1,mean))
  K_barFG = as.matrix(apply(K_FG,1,mean))
  K_barFZZ = as.matrix(apply(K_FZZ,1,mean))
  K_barGSH = as.matrix(apply(K_GSH,1,mean))
  K_barMBN= as.matrix(apply(K_MBN,1,mean))  
  
  
  dimension=max_time^2
  Cov_K_FG=Cov_K_MD=Cov_K_KC=Cov_K_FZZ=Cov_K_GSH=Cov_K_WL=Cov_K_MBN=matrix(0,dimension,dimension)
  for(i in 1:nsub)
  {
    Cov_K_MD = Cov_K_MD + (K_MD[,i]-K_barMD)%*%t(K_MD[,i]-K_barMD)/(nsub-1) #estimate for Omega
    Cov_K_KC = Cov_K_KC + (K_KC[,i]-K_barKC)%*%t(K_KC[,i]-K_barKC)/(nsub-1) #estimate for Omega
    Cov_K_FG = Cov_K_FG + (K_FG[,i]-K_barFG)%*%t(K_FG[,i]-K_barFG)/(nsub-1) #estimate for Omega
    Cov_K_FZZ = Cov_K_FZZ + (K_FZZ[,i]-K_barFZZ)%*%t(K_FZZ[,i]-K_barFZZ)/(nsub-1) #estimate for Omega
    Cov_K_GSH = Cov_K_GSH + (K_GSH[,i]-K_barGSH)%*%t(K_GSH[,i]-K_barGSH)/(nsub-1) #estimate for Omega
    Cov_K_MBN = Cov_K_MBN + (K_MBN[,i]-K_barMBN)%*%t(K_MBN[,i]-K_barMBN)/(nsub-1) #estimate for Omega
    
  } 
  
  
  #Obtaining the CENTER of the estimated covariance of the vec of the meat 
  dimension=max_time^2
  Center_FG = Center_MD = Center_KC = Center_FZZ = Center_GSH = Center_WL = Center_MBN = matrix(0,dimension,dimension) 

  loc1 <- 0
  loc2 <- 0
  for(i in 1:nsub){
   loc1 <- loc2 + 1
   loc2 <- loc1 + nobs[i] - 1
   yi <- as.matrix(y[loc1:loc2, ])
   xi <- x[loc1:loc2, ]
   ni <- nrow(yi)
   ti <- as.matrix(time[loc1:loc2, ])


   if (fam == "gaussian") 
    A_i = diag(max_time)

   if (fam == "poisson") 
    A_i = diag(as.vector(ui))


   if (fam == "gamma") 
    A_i = diag(as.vector(ui)) %*% diag(as.vector(ui))

   if (fam == "binomial") 
    A_i = diag(as.vector(ui)) %*% diag(as.vector(1 - ui))

   F_i = kronecker(sqrt(A_i),sqrt(A_i))

   Center_MD = Center_MD + solve(F_i)%*%Cov_K_MD%*%solve(F_i)/(nsub^2)
   Center_KC = Center_KC + solve(F_i)%*%Cov_K_KC%*%solve(F_i)/(nsub^2)
   Center_FG = Center_FG + solve(F_i)%*%Cov_K_FG%*%solve(F_i)/(nsub^2)
   Center_FZZ = Center_FZZ + solve(F_i)%*%Cov_K_FZZ%*%solve(F_i)/(nsub^2)
   Center_GSH = Center_GSH + solve(F_i)%*%Cov_K_GSH%*%solve(F_i)/(nsub^2)
   Center_MBN = Center_MBN + solve(F_i)%*%Cov_K_MBN%*%solve(F_i)/(nsub^2)     
  } #end of the for loop 
  

 #Obtaining the estimated covariance of the vec of the meat
 dimension=np^2
 Cov_vecM_MD = Cov_vecM_KC = Cov_vecM_FG = Cov_vecM_FZZ = Cov_vecM_WL = Cov_vecM_GSH = Cov_vecM_MBN =matrix(0,dimension,dimension)
 
  loc1 <- 0
  loc2 <- 0
  for(i in 1:nsub){
   loc1 <- loc2 + 1
   loc2 <- loc1 + nobs[i] - 1
   yi <- as.matrix(y[loc1:loc2, ])
   xi <- x[loc1:loc2, ]
   ni <- nrow(yi)
   ti <- as.matrix(time[loc1:loc2, ])


   #Note that I am obtaining the Pearson residuals in here, which are used to estimate the correlation parameter.
   if (fam == "gaussian") {            #Actually, it is probably a good idea not to use a guassian for this. I don't want to bother estimating the variance.
                ui <- xi %*% beta
                fui <- ui
                fui_dev <- diag(ni)
                vui <- diag(ni)
                Pearson = (yi-ui)     
   }
   if (fam == "poisson") {
                ui <- exp(xi %*% beta)
                fui <- log(ui)
                fui_dev <- diag(as.vector(ui))
                vui <- diag(as.vector(sqrt(1/ui)))
                Pearson = (yi-ui)/sqrt(ui)
   }
   if (fam == "gamma") {
                ui <- 1/(xi %*% beta)
                fui <- 1/ui
                fui_dev <- -diag(as.vector(ui)) %*% diag(as.vector(ui))
                vui <- diag(as.vector(1/ui))
                Pearson = (yi-ui)/ui
   }
   if (fam == "binomial") {
                ui <- 1/(1 + exp(-xi %*% beta))
                fui <- log(ui) - log(1 - ui)
                fui_dev <- diag(as.vector(ui)) %*% diag(as.vector(1 - ui))
                vui <- diag(as.vector(sqrt(1/ui))) %*% diag(as.vector(sqrt(1/(1 - ui))))
                Pearson = (yi-ui)/sqrt(ui*(1-ui))
   }






   if (corstr == "independence" ) {
                wi <- t(xi) %*% fui_dev %*% vui %*% vui
                Di = fui_dev %*% xi
   }


   if (corstr == "unstructured") {
	       Ri_i=diag(ni)
               for(k in 1:(ni-1))
                for(l in (k+1):ni)                
                 Ri_i[k,l]=Ri_i[l,k]=Ri_max[(ti[k,1]),(ti[l,1])]

               Ri_inv = solve(Ri_i)         
               wi <- t(xi) %*% fui_dev %*% vui %*% Ri_inv %*% vui
               Di = fui_dev %*% xi
   }


   if (corstr == "exchangeable") {
               Ri = diag(ni) + matrix(corr_par,ni,ni) - diag(corr_par,ni)     #This is the current estimated working correlation matrix
               Ri_inv = solve(Ri)           
               wi <- t(xi) %*% fui_dev %*% vui %*% Ri_inv %*% vui
               Di = fui_dev %*% xi
   }
   

   if (corstr == "AR-1") {
               Ri = matrix(0,ni,ni)                          #This is the current estimated working correlation matrix
               for(k in 1:ni)
                for(l in 1:ni)
                 Ri[k,l]=corr_par^abs(ti[k,1]-ti[l,1])  

               Ri_inv = solve(Ri)           
               wi <- t(xi) %*% fui_dev %*% vui %*% Ri_inv %*% vui
               Di = fui_dev %*% xi
   }
   

   if (corstr == "MA-1") {
               Ri = diag(ni)                          #This is the current estimated working correlation matrix
               for(k in 1:(ni-1))
                if( (ti[(k+1),1]-ti[k,1])==1 )
                 Ri[k,(k+1)]=Ri[(k+1),k]=corr_par

               Ri_inv = solve(Ri)           
               wi <- t(xi) %*% fui_dev %*% vui %*% Ri_inv %*% vui
               Di = fui_dev %*% xi
   }
   

   if (corstr == "Toeplitz") {
               Ri = diag(ni)                          #This is the current estimated working correlation matrix
               for(j in 1:(ni-1))             
                for(k in (j+1):ni)
                 {
                  position = (ti[k,1]-ti[j,1])
                  Ri[k,j]=Ri[j,k]=corr_par_matrix[position,1]
                 }

               Ri_inv = solve(Ri)           
               wi <- t(xi) %*% fui_dev %*% vui %*% Ri_inv %*% vui
               Di = fui_dev %*% xi
   }


   if (fam == "gaussian") 
    A_i = diag(max_time)

   if (fam == "poisson") 
    A_i = diag(as.vector(ui))


   if (fam == "gamma") 
    A_i = diag(as.vector(ui)) %*% diag(as.vector(ui))

   if (fam == "binomial") 
    A_i = diag(as.vector(ui)) %*% diag(as.vector(1 - ui))

   F_i = kronecker(sqrt(A_i),sqrt(A_i))

   C_i_init = kronecker(wi,wi)
   C_i = matrix(0,np*np,max_time*max_time)
   C_i[1:(np*np),1:(ni*ni)] = C_i_init   
     
   Cov_vecM_MD = Cov_vecM_MD + C_i %*% F_i %*% Center_MD %*% F_i %*% t(C_i)
   Cov_vecM_KC = Cov_vecM_KC + C_i %*% F_i %*% Center_KC %*% F_i %*% t(C_i) 
   Cov_vecM_FG = Cov_vecM_FG + C_i %*% F_i %*% Center_FG %*% F_i %*% t(C_i) 
   Cov_vecM_FZZ = Cov_vecM_FZZ + C_i %*% F_i %*% Center_FZZ %*% F_i %*% t(C_i) 
   Cov_vecM_GSH = Cov_vecM_GSH + C_i %*% F_i %*% Center_GSH %*% F_i %*% t(C_i) 
   Cov_vecM_WL = Cov_vecM_WL + C_i %*% F_i %*% Center_WL %*% F_i %*% t(C_i) 
   Cov_vecM_MBN = Cov_vecM_MBN + C_i %*% F_i %*% Center_MBN %*% F_i %*% t(C_i)     
   
   
  } #end of the for loop 


 #Obtaining the estimated covariance of the vec of my proposed covariance estimator (second to last formula at the bottom of page 1 of my derivation)

 V_M = (diag(np)+G) %*% solve(I)
 left_bun = kronecker(V_M,V_M)
 right_bun = t(kronecker(V_M,V_M))

 Cov_vecWB_MD = left_bun %*% Cov_vecM_MD %*% right_bun
 Cov_vecWB_KC = left_bun %*% Cov_vecM_KC %*% right_bun
 Cov_vecWB_FG = left_bun %*% Cov_vecM_FG %*% right_bun
 Cov_vecWB_FZZ = left_bun %*% Cov_vecM_FZZ %*% right_bun  
 Cov_vecWB_GSH = left_bun %*% Cov_vecM_GSH %*% right_bun  
 Cov_vecWB_WL = left_bun %*% Cov_vecM_WL %*% right_bun
 Cov_vecWB_MBN = left_bun %*% Cov_vecM_MBN %*% right_bun
 



 #################################For new df based on Pan (2001) formulas 
 #Obtaining the tau for the df

 #Estimated variances of the estimated variances of the estimated betas (regression np)
 tau_MD_Unst_New = tau_KC_Unst_New = tau_FG_Unst_New = tau_FZZ_Unst_New = tau_WL_Unst_New = tau_GSH_Unst_New = 
   tau_MBN_Unst_New =matrix(0,np,1)
 loc=1
 for(j in 1:np)
 {
   tau_MD_Unst_New[j,1]=Cov_vecWB_MD[loc,loc]  
   tau_KC_Unst_New[j,1]=Cov_vecWB_KC[loc,loc] 
   tau_FG_Unst_New[j,1]=Cov_vecWB_FG[loc,loc]
   tau_FZZ_Unst_New[j,1]=Cov_vecWB_FZZ[loc,loc]
   tau_GSH_Unst_New[j,1]=Cov_vecWB_GSH[loc,loc]
   tau_WL_Unst_New[j,1]=Cov_vecWB_WL[loc,loc]
   tau_MBN_Unst_New[j,1]=Cov_vecWB_MBN[loc,loc]
   loc = loc + np + 1
 }
 











































 #Obtaining the means of P for the Pan and Wall df
 P_barMD = as.matrix(apply(P_matrixMD,1,mean))
 P_barKC = as.matrix(apply(P_matrixKC,1,mean))
 P_barFG = as.matrix(apply(P_matrixFG,1,mean))  
 P_barFZZ = as.matrix(apply(P_matrixFZZ,1,mean))  
 P_barGSH = as.matrix(apply(P_matrixGSH,1,mean))  
 P_barMBN = as.matrix(apply(P_matrixMBN,1,mean))  
 
 
 dimension=np^2
 T_MD = T_KC = T_FG = T_FZZ = T_GSH = T_WL = T_MBN =matrix(0,dimension,dimension)
 for(i in 1:nsub)
 {
   T_MD = T_MD + (P_matrixMD[,i]-P_barMD)%*%t(P_matrixMD[,i]-P_barMD)/nsub/(nsub-1)
   T_KC = T_KC + (P_matrixKC[,i]-P_barKC)%*%t(P_matrixKC[,i]-P_barKC)/nsub/(nsub-1)
   T_FG = T_KC + (P_matrixFG[,i]-P_barFG)%*%t(P_matrixFG[,i]-P_barFG)/nsub/(nsub-1)
   T_FZZ = T_FZZ + (P_matrixFZZ[,i]-P_barFZZ)%*%t(P_matrixFZZ[,i]-P_barFZZ)/nsub/(nsub-1)
   T_GSH = T_GSH + (P_matrixGSH[,i]-P_barGSH)%*%t(P_matrixGSH[,i]-P_barGSH)/nsub/(nsub-1)
   T_MBN = T_MBN + (P_matrixMBN[,i]-P_barMBN)%*%t(P_matrixMBN[,i]-P_barMBN)/nsub/(nsub-1)
 } 
 
 V_M = (diag(np)+G) %*% solve(I)
 
 cov_vector_MD = nsub^2*kronecker(V_M,V_M)%*%T_MD%*%t(kronecker(V_M,V_M))
 cov_vector_KC = nsub^2*kronecker(V_M,V_M)%*%T_KC%*%t(kronecker(V_M,V_M))
 cov_vector_FG = nsub^2*kronecker(V_M,V_M)%*%T_FG%*%t(kronecker(V_M,V_M))
 cov_vector_FZZ = nsub^2*kronecker(V_M,V_M)%*%T_FZZ%*%t(kronecker(V_M,V_M))  
 cov_vector_GSH = nsub^2*kronecker(V_M,V_M)%*%T_GSH%*%t(kronecker(V_M,V_M))  
 cov_vector_MBN = nsub^2*kronecker(V_M,V_M)%*%T_MBN%*%t(kronecker(V_M,V_M))  
 

 tau_MD = tau_KC = tau_FG = tau_FZZ =tau_GSH = tau_MBN =matrix(0,np,1)
 loc=1
 for(j in 1:np)
 {
   tau_MD[j,1]=cov_vector_MD[loc,loc]  
   tau_FZZ[j,1]=cov_vector_FZZ[loc,loc]  
   tau_KC[j,1]=cov_vector_KC[loc,loc] 
   tau_FG[j,1]=cov_vector_FG[loc,loc] 
   tau_GSH[j,1]=cov_vector_GSH[loc,loc]
   tau_MBN[j,1]=cov_vector_MBN[loc,loc] 
   loc = loc + np + 1
 }
 
 
 #In very rare situations negative estimates can result. Checking and correcting for that here.  
 negative1=negative2=negative3=negative4=negative5=negative6=negative7=0
 for(j in 1:np)
 {
   if(tau_MD[j,1]<0) negative1=1  
   if(tau_KC[j,1]<0) negative2=1 
   if(tau_FG[j,1]<0) negative3=1 
   if(tau_FZZ[j,1]<0) negative4=1  
   if(tau_GSH[j,1]<0) negative5=1  
   if(tau_MBN[j,1]<0) negative7=1  
 }
 
 if(negative1==1)
 {
   V_M = solve(I)
   cov_vector_MD = nsub^2*kronecker(V_M,V_M)%*%T_MD%*%t(kronecker(V_M,V_M))
   tau_MD = matrix(0,np,1)
   loc=1
   for(j in 1:np)
   {
     tau_MD[j,1]=cov_vector_MD[loc,loc]     
     loc = loc + np + 1
   }
 }
 
 if(negative2==1)
 {
   V_M = solve(I)
   cov_vector_KC = nsub^2*kronecker(V_M,V_M)%*%T_KC%*%t(kronecker(V_M,V_M))
   tau_KC = matrix(0,np,1)
   loc=1
   for(j in 1:np)
   {
     tau_KC[j,1]=cov_vector_KC[loc,loc]     
     loc = loc + np + 1
   }
 }
 
 if(negative3==1)
 {
   V_M = solve(I)
   cov_vector_FG = nsub^2*kronecker(V_M,V_M)%*%T_FG%*%t(kronecker(V_M,V_M))
   tau_FG = matrix(0,np,1)
   loc=1
   for(j in 1:np)
   {
     tau_FG[j,1]=cov_vector_FG[loc,loc]     
     loc = loc + np + 1
   }
 }  
 
 if(negative4==1)
 {
   V_M = solve(I)
   cov_vector_FZZ = nsub^2*kronecker(V_M,V_M)%*%T_FZZ%*%t(kronecker(V_M,V_M))
   tau_FZZ = matrix(0,np,1)
   loc=1
   for(j in 1:np)
   {
     tau_FZZ[j,1]=cov_vector_FZZ[loc,loc]     
     loc = loc + np + 1
   }
 }
 
 if(negative5==1)
 {
   V_M = solve(I)
   cov_vector_GSH = nsub^2*kronecker(V_M,V_M)%*%T_GSH%*%t(kronecker(V_M,V_M))
   tau_GSH = matrix(0,np,1)
   loc=1
   for(j in 1:np)
   {
     tau_GSH[j,1]=cov_vector_GSH[loc,loc]     
     loc = loc + np + 1
   }
 }
 

 if(negative7==1)
 {
   V_M = solve(I)
   cov_vector_MBN = nsub^2*kronecker(V_M,V_M)%*%T_MBN%*%t(kronecker(V_M,V_M))
   tau_MBN = matrix(0,np,1)
   loc=1
   for(j in 1:np)
   {
     tau_MBN[j,1]=cov_vector_MBN[loc,loc]     
     loc = loc + np + 1
   }
 }
 





  #Now to get the final unstructured components of the meat;  Wang and Long (2011)

 	       for(j in 1:max_time)
                for(k in 1:max_time) 
                {              
                 R_Unst_typical[j,k] = R_Unst_typical[j,k]/counting[j,k]
                 R_Unst_MD[j,k] = R_Unst_MD[j,k]/counting[j,k]
                }



  #Now to get the meats

  meat_Unst_typical=0              
  meat_Unst_MD=0              


  #Working on getting the df for Wang and Long (2011) and Westgate and Burchett (2016)
  P_matrixMD_Unst = P_matrixKC_Unst = P_matrixMD_UnstWB= P_matrixFG_Unst =NULL


  loc1 <- 0
  loc2 <- 0
  for(i in 1:nsub){
   loc1 <- loc2 + 1
   loc2 <- loc1 + nobs[i] - 1
   yi <- as.matrix(y[loc1:loc2, ])
   xi <- x[loc1:loc2, ]
   ni <- nrow(yi)
   ti <- as.matrix(time[loc1:loc2, ])


   #Note that I am obtaining the Pearson residuals in here, which are used to estimate the correlation parameter.
   if (fam == "gaussian") {            #Actually, it is probably a good idea not to use a guassian for this. I don't want to bother estimating the variance.
                ui <- xi %*% beta
                fui <- ui
                fui_dev <- diag(ni)
                vui <- diag(ni)
                Ai_half = diag(ni)
                Pearson = (yi-ui)     
   }
   if (fam == "poisson") {
                ui <- exp(xi %*% beta)
                fui <- log(ui)
                fui_dev <- diag(as.vector(ui))
                vui <- diag(as.vector(sqrt(1/ui)))
                Ai_half = diag(as.vector(sqrt(ui)))
                Pearson = (yi-ui)/sqrt(ui)
   }
   if (fam == "gamma") {
                ui <- 1/(xi %*% beta)
                fui <- 1/ui
                fui_dev <- -diag(as.vector(ui)) %*% diag(as.vector(ui))
                vui <- diag(as.vector(1/ui))
                Ai_half = diag(as.vector(ui))
                Pearson = (yi-ui)/ui
   }
   if (fam == "binomial") {
                ui <- 1/(1 + exp(-xi %*% beta))
                fui <- log(ui) - log(1 - ui)
                fui_dev <- diag(as.vector(ui)) %*% diag(as.vector(1 - ui))
                vui <- diag(as.vector(sqrt(1/ui))) %*% diag(as.vector(sqrt(1/(1 - ui))))
                Ai_half = diag(as.vector(sqrt(ui))) %*% diag(as.vector(sqrt((1 - ui))))
                Pearson = (yi-ui)/sqrt(ui*(1-ui))
   }






   if (corstr == "independence" ) {
                wi <- t(xi) %*% fui_dev %*% vui %*% vui
                Di = fui_dev %*% xi
   }


   if (corstr == "unstructured") {
	       Ri_i=diag(ni)
               for(k in 1:(ni-1))
                for(l in (k+1):ni)                
                 Ri_i[k,l]=Ri_i[l,k]=Ri_max[(ti[k,1]),(ti[l,1])]

               Ri_inv = solve(Ri_i)         
               wi <- t(xi) %*% fui_dev %*% vui %*% Ri_inv %*% vui
               Di = fui_dev %*% xi
   }


   if (corstr == "exchangeable") {
               Ri = diag(ni) + matrix(corr_par,ni,ni) - diag(corr_par,ni)   
               Ri_inv = solve(Ri)           
               wi <- t(xi) %*% fui_dev %*% vui %*% Ri_inv %*% vui
               Di = fui_dev %*% xi
   }
   

   if (corstr == "AR-1") {
               Ri = matrix(0,ni,ni)                       
               for(k in 1:ni)
                for(l in 1:ni)
                 Ri[k,l]=corr_par^abs(ti[k,1]-ti[l,1])  

               Ri_inv = solve(Ri)           
               wi <- t(xi) %*% fui_dev %*% vui %*% Ri_inv %*% vui
               Di = fui_dev %*% xi
   }
   

   if (corstr == "MA-1") {
               Ri = diag(ni)                      
               for(k in 1:(ni-1))
                if( (ti[(k+1),1]-ti[k,1])==1 )
                 Ri[k,(k+1)]=Ri[(k+1),k]=corr_par

               Ri_inv = solve(Ri)           
               wi <- t(xi) %*% fui_dev %*% vui %*% Ri_inv %*% vui
               Di = fui_dev %*% xi
   }
   

   if (corstr == "Toeplitz") {
               Ri = diag(ni)                      
               for(j in 1:(ni-1))             
                for(k in (j+1):ni)
                 {
                  position = (ti[k,1]-ti[j,1])
                  Ri[k,j]=Ri[j,k]=corr_par_matrix[position,1]
                 }

               Ri_inv = solve(Ri)           
               wi <- t(xi) %*% fui_dev %*% vui %*% Ri_inv %*% vui
               Di = fui_dev %*% xi
   }


  
           middle_Unst_typical = middle_Unst_MD = middle_Unst_KC = matrix(0,ni,ni)
    	       for(j in 1:ni)
                for(k in 1:ni) 
                {              
                 middle_Unst_typical[j,k] = R_Unst_typical[(ti[j,1]),(ti[k,1])]
                 middle_Unst_MD[j,k] = R_Unst_MD[(ti[j,1]),(ti[k,1])]
                }  

  meat_Unst_typical = meat_Unst_typical + wi %*% Ai_half %*% middle_Unst_typical %*% Ai_half %*% t(wi)      
  meat_Unst_MD = meat_Unst_MD + wi %*% Ai_half %*% middle_Unst_MD %*% Ai_half %*% t(wi)              

  #Obtaining the matrices needed to calculate the Wang and Long (2011)
  P_MD_Unst_i = P_KC_Unst_i = NULL
  meat_MD_Unst_i = wi %*% Ai_half %*% R_Unst_MD[1:ni,1:ni] %*% Ai_half %*% t(wi) 

  
  for(j in 1:np)
  {
    P_MD_Unst_i=c(P_MD_Unst_i,meat_MD_Unst_i[,j])

  }
  
  P_MD_Unst_i=as.matrix(P_MD_Unst_i)
  
  
  P_matrixMD_Unst = cbind(P_matrixMD_Unst,P_MD_Unst_i)


  } #end of the for loop (so done adding up the GEE and I and residual terms for each independent subject)



















  loc1j <- 0
  loc2j <- 0
  for(j in 1:nsub){
   loc1j <- loc2j + 1
   loc2j <- loc1j + nobs[j] - 1
   yj <- as.matrix(y[loc1j:loc2j, ])
   xj <- x[loc1j:loc2j, ]
   nj <- nrow(yj)
   tj <- as.matrix(time[loc1j:loc2j, ])


   #Note that I am obtaining the Pearson residuals in here, which are used to estimate the correlation parameter.
   if (fam == "gaussian") {         
                uj <- xj %*% beta
                fuj <- uj
                fuj_dev <- diag(nj)
                vuj <- diag(nj)
                Aj_half = diag(nj)
                Pearson = (yj-uj)     
   }
   if (fam == "poisson") {
                uj <- exp(xj %*% beta)
                fuj <- log(uj)
                fuj_dev <- diag(as.vector(uj))
                vuj <- diag(as.vector(sqrt(1/uj)))
                Aj_half = diag(as.vector(sqrt(uj)))
                Pearson = (yj-uj)/sqrt(uj)
   }
   if (fam == "gamma") {
                uj <- 1/(xj %*% beta)
                fuj <- 1/uj
                fuj_dev <- -diag(as.vector(uj)) %*% diag(as.vector(uj))
                vuj <- diag(as.vector(1/uj))
                Aj_half = diag(as.vector(uj))
                Pearson = (yj-uj)/uj
   }
   if (fam == "binomial") {
                uj <- 1/(1 + exp(-xj %*% beta))
                fuj <- log(uj) - log(1 - uj)
                fuj_dev <- diag(as.vector(uj)) %*% diag(as.vector(1 - uj))
                vuj <- diag(as.vector(sqrt(1/uj))) %*% diag(as.vector(sqrt(1/(1 - uj))))
                Aj_half = diag(as.vector(sqrt(uj))) %*% diag(as.vector(sqrt((1 - uj))))
                Pearson = (yj-uj)/sqrt(uj*(1-uj))
   }






   if (corstr == "independence" ) {
                wj <- t(xj) %*% fuj_dev %*% vuj %*% vuj
                Dj = fuj_dev %*% xj
   }


   if (corstr == "unstructured") {
	       Rj_i=diag(nj)
               for(k in 1:(nj-1))
                for(l in (k+1):nj)                
                 Rj_i[k,l]=Rj_i[l,k]=Ri_max[(tj[k,1]),(tj[l,1])]

               Rj_inv = solve(Rj_i)         
               wj <- t(xj) %*% fuj_dev %*% vuj %*% Rj_inv %*% vuj
               Dj = fuj_dev %*% xj
   }


   if (corstr == "exchangeable") {
               Rj = diag(nj) + matrix(corr_par,nj,nj) - diag(corr_par,nj)     
               Rj_inv = solve(Rj)           
               wj <- t(xj) %*% fuj_dev %*% vuj %*% Rj_inv %*% vuj
               Dj = fuj_dev %*% xj
   }
   

   if (corstr == "AR-1") {
               Rj = matrix(0,nj,nj)             
               for(k in 1:nj)
                for(l in 1:nj)
                 Rj[k,l]=corr_par^abs(tj[k,1]-tj[l,1])  

               Rj_inv = solve(Rj)           
               wj <- t(xj) %*% fuj_dev %*% vuj %*% Rj_inv %*% vuj
               Dj = fuj_dev %*% xj
   }
   

   if (corstr == "MA-1") {
               Rj = diag(nj)                    
               for(k in 1:(nj-1))
                if( (tj[(k+1),1]-tj[k,1])==1 )
                 Rj[k,(k+1)]=Rj[(k+1),k]=corr_par

               Rj_inv = solve(Rj)           
               wj <- t(xj) %*% fuj_dev %*% vuj %*% Rj_inv %*% vuj
               Dj = fuj_dev %*% xj
   }
   

   if (corstr == "Toeplitz") {
               Rj = diag(nj)                    
               for(l in 1:(nj-1))             
                for(k in (l+1):nj)
                 {
                  position = (tj[k,1]-tj[l,1])
                  Rj[k,l]=Rj[l,k]=corr_par_matrix[position,1]
                 }

               Rj_inv = solve(Rj)           
               wj <- t(xj) %*% fuj_dev %*% vuj %*% Rj_inv %*% vuj
               Dj = fuj_dev %*% xj
   }

  
  I_j = diag(nj)
  H_jj = Dj%*%(diag(np)+G)%*%solve(I)%*%wj
  middle_meatMD_j = matrix(0,nj,nj)
  middle_meatMD_j[1:nj,1:nj] = vuj %*% solve(I_j-H_jj) %*% (yj-uj) %*% t(yj-uj) %*% solve(I_j-t(H_jj)) %*% vuj

  meat_MD_Unst_j = 0
  loc1 <- 0
  loc2 <- 0
  for(i in 1:nsub){
   loc1 <- loc2 + 1
   loc2 <- loc1 + nobs[i] - 1
   yi <- as.matrix(y[loc1:loc2, ])
   xi <- x[loc1:loc2, ]
   ni <- nrow(yi)
   ti <- as.matrix(time[loc1:loc2, ])


   if (fam == "gaussian") {       
                ui <- xi %*% beta
                fui <- ui
                fui_dev <- diag(ni)
                vui <- diag(ni)
                Ai_half = diag(ni)
                Pearson = (yi-ui)     
   }
   if (fam == "poisson") {
                ui <- exp(xi %*% beta)
                fui <- log(ui)
                fui_dev <- diag(as.vector(ui))
                vui <- diag(as.vector(sqrt(1/ui)))
                Ai_half = diag(as.vector(sqrt(ui)))
                Pearson = (yi-ui)/sqrt(ui)
   }
   if (fam == "gamma") {
                ui <- 1/(xi %*% beta)
                fui <- 1/ui
                fui_dev <- -diag(as.vector(ui)) %*% diag(as.vector(ui))
                vui <- diag(as.vector(1/ui))
                Ai_half = diag(as.vector(ui))
                Pearson = (yi-ui)/ui
   }
   if (fam == "binomial") {
                ui <- 1/(1 + exp(-xi %*% beta))
                fui <- log(ui) - log(1 - ui)
                fui_dev <- diag(as.vector(ui)) %*% diag(as.vector(1 - ui))
                vui <- diag(as.vector(sqrt(1/ui))) %*% diag(as.vector(sqrt(1/(1 - ui))))
                Ai_half = diag(as.vector(sqrt(ui))) %*% diag(as.vector(sqrt((1 - ui))))
                Pearson = (yi-ui)/sqrt(ui*(1-ui))
   }






   if (corstr == "independence" ) {
                wi <- t(xi) %*% fui_dev %*% vui %*% vui
                Di = fui_dev %*% xi
   }


   if (corstr == "unstructured") {
	       Ri_i=diag(ni)
               for(k in 1:(ni-1))
                for(l in (k+1):ni)                
                 Ri_i[k,l]=Ri_i[l,k]=Ri_max[(ti[k,1]),(ti[l,1])]

               Ri_inv = solve(Ri_i)         
               wi <- t(xi) %*% fui_dev %*% vui %*% Ri_inv %*% vui
               Di = fui_dev %*% xi
   }


   if (corstr == "exchangeable") {
               Ri = diag(ni) + matrix(corr_par,ni,ni) - diag(corr_par,ni)     
               Ri_inv = solve(Ri)           
               wi <- t(xi) %*% fui_dev %*% vui %*% Ri_inv %*% vui
               Di = fui_dev %*% xi
   }
   

   if (corstr == "AR-1") {
               Ri = matrix(0,ni,ni)                     
               for(k in 1:ni)
                for(l in 1:ni)
                 Ri[k,l]=corr_par^abs(ti[k,1]-ti[l,1])  

               Ri_inv = solve(Ri)           
               wi <- t(xi) %*% fui_dev %*% vui %*% Ri_inv %*% vui
               Di = fui_dev %*% xi
   }
   

   if (corstr == "MA-1") {
               Ri = diag(ni)                      
               for(k in 1:(ni-1))
                if( (ti[(k+1),1]-ti[k,1])==1 )
                 Ri[k,(k+1)]=Ri[(k+1),k]=corr_par

               Ri_inv = solve(Ri)           
               wi <- t(xi) %*% fui_dev %*% vui %*% Ri_inv %*% vui
               Di = fui_dev %*% xi
   }
   

   if (corstr == "Toeplitz") {
               Ri = diag(ni)                      
               for(l in 1:(ni-1))             
                for(k in (l+1):ni)
                 {
                  position = (ti[k,1]-ti[l,1])
                  Ri[k,l]=Ri[l,k]=corr_par_matrix[position,1]
                 }

               Ri_inv = solve(Ri)           
               wi <- t(xi) %*% fui_dev %*% vui %*% Ri_inv %*% vui
               Di = fui_dev %*% xi
   }



  if (ni <= nj)  meat_MD_Unst_j = meat_MD_Unst_j + wi %*% Ai_half %*% middle_meatMD_j[1:ni,1:ni] %*% Ai_half %*% t(wi)/nsub
  if (ni > nj) meat_MD_Unst_j = meat_MD_Unst_j + wi[,1:nj] %*% Ai_half[1:nj,1:nj] %*% middle_meatMD_j[1:nj,1:nj] %*% Ai_half[1:nj,1:nj] %*% t(wi[,1:nj])/nsub
   
}

  

   #Obtaining the matrices needed to calculate Westgate and Burchett (2016) df 
   P_MD_Unst_j = NULL

   for(l in 1:np)
   {
     P_MD_Unst_j=c(P_MD_Unst_j,meat_MD_Unst_j[,l])
   }
   
   P_MD_Unst_j=as.matrix(P_MD_Unst_j)
   
   P_matrixMD_UnstWB = cbind(P_matrixMD_UnstWB,P_MD_Unst_j)
  
}


  #Obtaining the means of P
  P_barMD_Unst = as.matrix(apply(P_matrixMD_Unst,1,mean))
  P_barMD_UnstWB = as.matrix(apply(P_matrixMD_UnstWB,1,mean))

  
  
  dimension=np^2
  T_MD_Unst = T_MD_UnstWB = T_KC_Unst = T_FG_Unst = matrix(0,dimension,dimension)
  for(i in 1:nsub)
  {
    T_MD_Unst = T_MD_Unst + (P_matrixMD_Unst[,i]-P_barMD_Unst)%*%t(P_matrixMD_Unst[,i]-P_barMD_Unst)/nsub/(nsub-1)
    T_MD_UnstWB = T_MD_UnstWB + (P_matrixMD_UnstWB[,i]-P_barMD_UnstWB)%*%t(P_matrixMD_UnstWB[,i]-P_barMD_UnstWB)/nsub/(nsub-1)
    
    
    } 
  
  V_M = (diag(np)+G) %*% solve(I)
  cov_vector_MD_Unst = nsub^2*kronecker(V_M,V_M)%*%T_MD_Unst%*%t(kronecker(V_M,V_M))
  cov_vector_MD_UnstWB = nsub^2*kronecker(V_M,V_M)%*%T_MD_UnstWB%*%t(kronecker(V_M,V_M))

  #Estimated variances of the estimated variances of the estimated betas (regression np)
  tau_MD_Unst = tau_MD_UnstWB =  matrix(0,np,1)
  loc=1
  for(j in 1:np)
  {
    tau_MD_Unst[j,1]=cov_vector_MD_Unst[loc,loc]  
    tau_MD_UnstWB[j,1]=cov_vector_MD_UnstWB[loc,loc]  
    loc = loc + np + 1
  }
  
  
  
  
  negative1 = 0
  for(j in 1:np)
  {
    if(tau_MD_Unst[j,1]<0) negative1=1  
  }
  
  if(negative1==1)
  {
    V_M = solve(I)
    cov_vector_MD_Unst = nsub^2*kronecker(V_M,V_M)%*%T_MD_Unst%*%t(kronecker(V_M,V_M))
    tau_MD_Unst = matrix(0,np,1)
    loc=1
    for(j in 1:np)
    {
      tau_MD_Unst[j,1]=cov_vector_MD_Unst[loc,loc]     
      loc = loc + np + 1
    }
  }
  



  emp_BCMD = solve(I) %*% meat_BCMD %*% solve(I)
  emp_FZZ = solve(I) %*% meat_FZZ %*% solve(I)
  emp_BCKC = solve(I) %*% meat_BCKC %*% solve(I)
  emp_BCFG = solve(I) %*% meat_BCFG %*% solve(I)
  emp_MBN = solve(I) %*% meat_BCMBN %*% solve(I)
  emp_GST = solve(I) %*% meat_BCGST %*% solve(I)


  #With Westgate's (2016) correction factor, for information purposes only
  emp_BCMD_correction = (diag(np)+G) %*% solve(I) %*% meat_BCMD_correction %*% solve(I) %*% t(diag(np)+G)
  emp_FZZ_correction = (diag(np)+G) %*% solve(I) %*% meat_FZZ_correction %*% solve(I) %*% t(diag(np)+G)
  emp_BCKC_correction = (diag(np)+G) %*% solve(I) %*% meat_BCKC_correction %*% solve(I) %*% t(diag(np)+G)
  emp_BCFG_correction = (diag(np)+G) %*% solve(I) %*% meat_BCFG_correction %*% solve(I) %*% t(diag(np)+G)
  emp_GST_correction = (diag(np)+G) %*% emp_GST %*% t(diag(np)+G)
  emp_MBN_correction = (diag(np)+G) %*% emp_MBN %*% t(diag(np)+G)
  emp_correction = (diag(np)+G) %*% solve(I) %*% meat_emp_correction %*% solve(I) %*% t(diag(np)+G)
  
  
  emp_BCMD_correction_Unst = (diag(np)+G) %*% solve(I) %*% meat_Unst_MD %*% solve(I) %*% t(diag(np)+G)
  emp_correction_Unst = (diag(np)+G) %*% solve(I) %*% meat_Unst_typical %*% solve(I) %*% t(diag(np)+G)
  
  
  AVG_KCMD = (emp_BCMD + emp_BCKC)/2
  AVG_KCMD_correction = (emp_BCMD_correction + emp_BCKC_correction)/2



  for(k in 1:length(beta))                  #So for each beta
  {
    if(Model_basedVar[k,k]<=0) Model_basedVar=emp
    if(emp_BCKC[k,k]<=0) emp_BCKC=emp
    if(emp_BCMD[k,k]<=0) emp_BCMD=emp
    if(emp_FZZ[k,k]<=0) emp_FZZ=emp
    if(emp_GST[k,k]<=0) emp_GST=emp
    if(AVG_KCMD[k,k]<=0) AVG_KCMD=emp
    if(emp_BCFG[k,k]<=0) emp_BCFG=emp
    
    #Outputting with correction by Westgate (2016) for information purposes only
    if(MB_correction[k,k]<=0) MB_correction=emp_correction
    if(emp_BCKC_correction[k,k]<=0) emp_BCKC_correction=emp_correction
    if(emp_BCMD_correction[k,k]<=0) emp_BCMD_correction=emp_correction
    if(emp_FZZ_correction[k,k]<=0) emp_FZZ_correction=emp_correction
    if(AVG_KCMD_correction[k,k]<=0) AVG_KCMD_correction=emp
    if(emp_BCFG_correction[k,k]<=0) emp_BCFG_correction=emp_correction
  }
  
  #Degrees of Freedom using the Pan & Wall method
  df_MD = df_KC = df_FZZ = df_WL = df_AVG = df_FG = df_GSH = df_MBN = df_FZZ = df_MBN =matrix(0,np,1)
  for(j in 1:np)
  {
    df_MD[j,1] = 2*(emp_BCMD[j,j])^2/tau_MD[j,1]
    df_KC[j,1] = 2*(emp_BCKC[j,j])^2/tau_KC[j,1]
    df_AVG[j,1] = (df_MD[j,1] + df_KC[j,1])/2
    df_FG[j,1] = 2*(emp_BCFG[j,j])^2/tau_FG[j,1]
    df_GSH[j,1] = 2*(emp_GST[j,j])^2/tau_GSH[j,1]
    #Rare instances of GSH giving negative degrees of freedom for interaction term in S3 continuous
    #Setting to MD PW df in this case
    if (df_GSH[j,1]<0) df_GSH[j,1]=df_MD[j,1]
    df_FZZ[j,1] = 2*(emp_FZZ_correction[j,j])^2/tau_FZZ[j,1]
    df_MBN[j,1] = 2*(emp_MBN[j,j])^2/tau_MBN[j,1]
    
  }
  
  #Degrees of Freedom using the Wang & Long (2011) method
  df_MD_Unst = matrix(0,np,1)
  for(j in 1:np)
  {
    df_MD_Unst[j,1] = 2*(emp_BCMD_correction_Unst[j,j])^2/tau_MD_Unst[j,1]
    if (df_MD_Unst[j,1]<0) df_MD_Unst[j,1]=df_MD[j,1] #if negative, revert to Pan and Wall df.
  }
  
  
  
  #################################For new df based on Wang and Log 
  #Obtaining the Degrees of Freedom using my new method 
  df_MD_Unst_WB = matrix(0,np,1)
  for(j in 1:np)
  {
    df_MD_Unst_WB[j,1] = 2*(emp_BCMD_correction_Unst[j,j])^2/tau_MD_UnstWB[j,1]
  }
  
  
  
  
  #Outputting the parameter estimates and corresponding stimated SEs
  
  output =  matrix(0,40,length(beta)) #The first column is for beta0, the second for beta1, etc. 
  output[1,] = t(beta)
  
 for(k in 1:length(beta))
 {
   output[2,k] = sqrt(Model_basedVar[k,k])
   output[3,k] = sqrt(emp[k,k])
   output[4,k] = sqrt(emp_BCKC[k,k])
   output[5,k] = sqrt(emp_BCMD[k,k])
   output[7,k] = sqrt(AVG_KCMD[k,k])
   output[9,k] = sqrt(emp_BCFG[k,k])
   output[10,k] = sqrt(emp_GST[k,k])
   output[11,k] = sqrt(emp_MBN[k,k])
   output[12,k] = sqrt(emp_FZZ[k,k])
   
#   output[13,k] = sqrt(MB_correction[k,k])
#   output[14,k] = sqrt(emp_correction[k,k])
#   output[15,k] = sqrt(emp_BCKC_correction[k,k])
#   output[16,k] = sqrt(emp_BCMD_correction[k,k])
#   output[18,k] = sqrt(AVG_KCMD_correction[k,k])
#   output[20,k] = sqrt(emp_BCFG_correction[k,k])
#   output[21,k] = sqrt(emp_GST_correction[k,k])
#   output[22,k] = sqrt(emp_MBN_correction[k,k])
#   output[23,k] = sqrt(emp_FZZ_correction[k,k])
 }

 


 output[24,1]=non_converge

 if ((corstr == "exchangeable") | (corstr == "AR-1") | (corstr == "MA-1"))
 {
  output[25,1]=corr_par
  output[26,2]=scale
 }


 output[27,]=t(df_KC) #Pan and Wall df, KC
 output[28,]=t(df_MD) # Pan and Wall df, MD
 output[30,]=t((df_KC + df_MD)/2) # Pan and Wall df, Average MD/KC
 output[32,]=t(df_FG) # Pan and Wall, FG
 output[33,]=t(df_GSH) #Pan and Wall, GSH
 output[34,]=t(df_MBN) # Pan and Wall, FZZ
 output[35,]=t(df_FZZ) # Pan and Wall, MBN
 
 
 output[36,]=t(df_MD_Unst) #Wang and Long, MD
 output[37,]=t(df_MD_Unst_WB) #Westgate and Burchette, MD
 

 return(output)
}
